import datetime as dt

import factory
import factory.fuzzy
from app_utils.testdata_factories import UserMainFactory

from django.utils.timezone import now

from ..models import InactivityPing, InactivityPingConfig, LeaveOfAbsence


class UserMainDefaultFactory(UserMainFactory):
    """A normal user with access to this app."""

    permissions__ = ["inactivity.basic_access"]


class UserMainManagerFactory(UserMainFactory):
    """A user who can manage loa requests."""

    permissions__ = ["inactivity.manage_leave"]


class LeaveOfAbsenceFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = LeaveOfAbsence

    user = factory.SubFactory(UserMainDefaultFactory)
    approver = factory.SubFactory(UserMainManagerFactory)
    start = factory.fuzzy.FuzzyDateTime(
        start_dt=now() - dt.timedelta(days=3), force_microsecond=0
    )
    end = factory.LazyAttribute(
        lambda obj: factory.fuzzy.FuzzyDateTime(
            start_dt=obj.start, end_dt=now() + dt.timedelta(days=3), force_microsecond=0
        ).fuzz()
    )
    notes = factory.Faker("sentence")


class InactivityPingConfigFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = InactivityPingConfig

    name = factory.Faker("city")
    days = 3
    text = factory.Faker("sentences")


class InactivityPingFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = InactivityPing

    config = factory.SubFactory(InactivityPingConfigFactory)
    timestamp = factory.LazyFunction(now)
    user = factory.SubFactory(UserMainDefaultFactory)



# Generated by Django 4.0.3 on 2022-03-03 13:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("authentication", "0017_remove_fleetup_permission"),
        ("inactivity", "0003_add_webhooks"),
    ]

    operations = [
        migrations.AlterField(
            model_name="inactivitypingconfig",
            name="groups",
            field=models.ManyToManyField(
                blank=True,
                help_text="Groups subject to the inactivity policy. If empty, applies to all groups.",
                related_name="+",
                to="auth.group",
            ),
        ),
        migrations.AlterField(
            model_name="inactivitypingconfig",
            name="states",
            field=models.ManyToManyField(
                blank=True,
                help_text="States subject to the inactivity policy. If empty, applies to all states.",
                related_name="+",
                to="authentication.state",
            ),
        ),
    ]



# Generated by Django 3.1.2 on 2021-01-17 17:44

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("authentication", "0017_remove_fleetup_permission"),
        ("auth", "0012_alter_user_first_name_max_length"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (("basic_access", "Can access this app"),),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="LeaveOfAbsence",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "start",
                    models.DateField(help_text="The start of the leave of absence."),
                ),
                (
                    "end",
                    models.DateField(
                        blank=True,
                        help_text="The end of the leave of absence. Leave blank for an indefinite leave.",
                        null=True,
                    ),
                ),
                ("notes", models.TextField(blank=True)),
                (
                    "approver",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="InactivityPingConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Internal name for the ping config. Must be unique.",
                        max_length=48,
                        unique=True,
                    ),
                ),
                (
                    "days",
                    models.PositiveIntegerField(
                        help_text="The number of days the user must be inactive."
                    ),
                ),
                (
                    "text",
                    models.TextField(
                        help_text="The text of the message or notification sent to the end user."
                    ),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Groups subject to the inactivity ping. If empty, applies to all groups.",
                        related_name="_inactivitypingconfig_groups_+",
                        to="auth.Group",
                    ),
                ),
                (
                    "states",
                    models.ManyToManyField(
                        blank=True,
                        help_text="States subject to the inactivity ping. If empty, applies to all states.",
                        related_name="_inactivitypingconfig_states_+",
                        to="authentication.State",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="InactivityPing",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField()),
                (
                    "config",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="inactivity.inactivitypingconfig",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
    ]



import datetime
import logging

import dhooks_lite
from celery import shared_task

from django.contrib.auth.models import User
from django.db.models import Q

from memberaudit.models import Character

from .models import InactivityPing, InactivityPingConfig, Webhook
from .utils import notify_user

logger = logging.getLogger(__name__)


@shared_task
def send_inactivity_ping(user_pk: int, config_pk: int):
    config = InactivityPingConfig.objects.get(pk=config_pk)
    user = User.objects.get(pk=user_pk)
    InactivityPing.objects.create(
        config=config, timestamp=datetime.datetime.utcnow(), user=user
    )
    for webhook in Webhook.objects.filter(
        Q(ping_configs=config) | Q(ping_configs=None), Q(is_active=True)
    ):
        if str(Webhook.NOTIFICATION_TYPE_INACTIVE_USER) in webhook.notification_types:
            if webhook.webhook_type == Webhook.WEBHOOK_TYPE_DISCORD:
                hook = dhooks_lite.Webhook(webhook.url)
                hook.execute(
                    "**%(user_name)s** has been inactive for **%(days)s** day(s) and has been notified according to **%(config_name)s**"
                    % {
                        "user_name": user.profile.main_character.character_name,
                        "days": config.days,
                        "config_name": config.name,
                    }
                )
    notify_user(user.pk, config.text)


@shared_task
def check_inactivity_for_user(user_pk: int):
    now = datetime.datetime.now(datetime.timezone.utc).date()
    user = User.objects.get(pk=user_pk)
    if (
        user.leaveofabsence_set.filter(
            Q(start__lt=now), Q(end=None) | Q(end__gt=now), ~Q(approver=None)
        ).count()
        == 0
    ):
        last_loa = (
            user.leaveofabsence_set.filter(Q(end__lt=now), ~Q(approver=None))
            .order_by("-end")
            .first()
        )
        for config in InactivityPingConfig.objects.all():
            if config.is_applicable_to(user):
                threshold_date = datetime.datetime.now(
                    datetime.timezone.utc
                ).date() - datetime.timedelta(days=config.days)
                registered = Character.objects.owned_by_user(user).exists()
                active = (
                    Character.objects.owned_by_user(user)
                    .filter(
                        Q(online_status__last_login__gt=threshold_date)
                        | Q(online_status__last_logout__gt=threshold_date),
                    )
                    .exists()
                )
                excused = last_loa and (
                    not last_loa.end or threshold_date < last_loa.end
                )
                pinged = InactivityPing.objects.filter(
                    user__pk=user_pk, config=config
                ).exists()
                if active:
                    InactivityPing.objects.filter(
                        user__pk=user_pk, config=config
                    ).delete()
                if not active and registered and not pinged and not excused:
                    send_inactivity_ping(user_pk, config.pk)


# Inactivity Task
@shared_task
def check_inactivity():
    for user in User.objects.all():
        check_inactivity_for_user(user.pk)



from app_utils.testdata_factories import UserFactory
from app_utils.testing import NoSocketsTestCase

from .factories import InactivityPingFactory


class TestInactivityPing(NoSocketsTestCase):
    def test_str_normal(self):
        # given
        ping = InactivityPingFactory()
        # when
        result = str(ping)
        # then
        self.assertIn(ping.user.last_name, result)

    def test_str_user_without_main(self):
        # given
        ping = InactivityPingFactory(user=UserFactory())
        # when
        result = str(ping)
        # then
        self.assertTrue(result)



from django.urls import path

from . import views

app_name = "inactivity"

urlpatterns = [
    path("", views.index, name="index"),
    path("manage", views.manage, name="manage_requests"),
    path("loa_requests", views.list_loa_requests, name="list_loa_requests"),
    path(
        "loa_requests/pending",
        views.list_pending_loa_requests,
        name="list_pending_loa_requests",
    ),
    path("loa_requests/create", views.create_loa_request, name="create_loa_request"),
    path(
        "loa_requests/<int:request_id>/cancel",
        views.cancel_loa_request,
        name="cancel_loa_request",
    ),
    path(
        "loa_requests/<int:request_id>/approve",
        views.approve_loa_request,
        name="approve_loa_request",
    ),
    path(
        "loa_requests/<int:request_id>",
        views.view_loa_request,
        name="view_loa_request",
    ),
]
